"""A wrapper around PyChef to deal with configuration"""

from __future__ import absolute_import

import urllib  # urllib.parse for Python 3

import chef.api
import chef.exceptions


class Chef(object):
    @classmethod
    def from_flask_app(cls, app):
        config_keys = ('CHEF_URL', 'CHEF_KEY', 'CHEF_CLIENT')

        for key in config_keys:
            if not key in app.config:
                raise Exception(key + " has not been set")

        return cls(*(app.config[x] for x in config_keys))

    def __init__(self, url, key, client):
        self.client = chef.api.ChefAPI(url, key, client)

    def _get(self, path):
        return self.client.api_request('GET', path)

    def get(self, *segments):
        return self._get('/' + '/'.join(segments))

    # Search API
    # GET /search

    def _search(self, index, query, keys=None, rows=1000, start=0):
        args = urllib.urlencode(dict(q=query, rows=rows, start=start))
        url = '/search/{0}?{1}'.format(index, args)
        method = 'POST' if keys is not None else 'GET'
        return self.client.api_request(method, url, data=keys)

    @staticmethod
    def _build_query(query, operator=' AND '):
        if isinstance(query, dict):
            return operator.join((k + ':' + v for k, v in query.items()))
        return query

    def search(self, index, query, rows=1000, start=0):
        return self._search(index, self._build_query(query), rows, start)

    # Partial Search API
    # POST /search

    @staticmethod
    def _build_partial_search(keys):
        """Create a partial_search from a list of attributes or a dict"""
        if isinstance(keys, (list, tuple)):
            keys = [key.split('.') for key in keys]
            keys = dict((key[0], key) for key in keys)
        return keys

    def partial_search(self, index, query, keys={}, rows=1000, start=0):
        """Returns data from a partial search (only returns defined keys)"""
        query = self._build_query(query)
        keys = self._build_partial_search(keys)

        # Sets a small number of very useful default attributes
        if index == 'node':
            keys.setdefault('name', ['name'])
            keys.setdefault('chef_environment', ['chef_environment'])

        # Returns data in this structure:
        # {rows: [{data: {...}, url: '...'}], start: 0, total: 1}
        results = self._search(index, query, keys, rows, start)

        # Return a list containing the returned data for each row
        return [row['data'] for row in results['rows']]



"""Flask blueprints for sous-chef"""

from __future__ import absolute_import

import flask
import chef.exceptions

__all__ = ['ui']

ui = flask.Blueprint('ui', __name__)


@ui.errorhandler(404)
@ui.errorhandler(chef.exceptions.ChefServerNotFoundError)
def not_found(error):
    return flask.render_template('404.html', error=error), 404


@ui.before_request
def get_chef_environments():
    """Ensure a list of environments is availible"""
    flask.g.chef_environments = flask.current_app.chef.get('environments')


@ui.url_value_preprocessor
def pop_chef_environment(endpoint, values):
    """Set the current environment, defaulting to * for all environments"""
    flask.g.chef_environment = values.pop('environment', '*')


@ui.url_defaults
def put_chef_environment(endpoint, values):
    """Set a default environment in calls to url_for"""
    expects = flask.current_app.url_map.is_endpoint_expecting
    if expects(endpoint, 'environment') and flask.g.chef_environment != '*':
        values.setdefault('environment', flask.g.chef_environment)


# Environments

@ui.route('/', endpoint='home')
@ui.route('/environments')
def environments():
    return flask.render_template(
        'environments.html', environments=flask.g.chef_environments)


@ui.route('/<string:environment>')
@ui.route('/environments/<string:environment>')
def environment():
    environment = flask.current_app.chef.get(
        'environments', flask.g.chef_environment)
    nodes = flask.current_app.chef.partial_search(
        'node', {'chef_environment': flask.g.chef_environment})
    return flask.render_template(
        'environment.html', environment=environment, nodes=nodes)


# Roles

@ui.route('/roles')
@ui.route('/<environment>/roles')
def roles():
    search = {'roles': '*', 'chef_environment': flask.g.chef_environment}
    nodes = flask.current_app.chef.partial_search('node', search, ['roles'])
    roles = set((r for node in nodes for r in node['roles']))
    return flask.render_template('roles.html', roles=roles)


@ui.route('/roles/<string:role>')
@ui.route('/<environment>/roles/<string:role>')
def role(role):
    _role = flask.current_app.chef.get('roles', role)
    nodes = flask.current_app.chef.partial_search('node', {
        'chef_environment': flask.g.chef_environment,
        'roles': role
    })
    return flask.render_template('role.html', role=_role, nodes=nodes)


# Nodes

@ui.route('/nodes')
@ui.route('/<environment>/nodes')
def nodes():
    nodes = flask.current_app.chef.partial_search('node', {
        'chef_environment': flask.g.chef_environment
    })
    return flask.render_template('nodes.html', nodes=nodes)


def get_node(name, keys={}):
    """Returns a single row or raises an error"""
    nodes = flask.current_app.chef.partial_search('node', {
        'chef_environment': flask.g.chef_environment,
        'name': name
    }, keys)

    if not nodes:
        raise chef.exceptions.ChefServerNotFoundError('No rows returned')
    elif len(nodes) > 1:
        raise chef.exceptions.ChefServerNotFoundError('Multiple rows returned')
    else:
        return nodes[0]


@ui.route('/nodes/<string:node>')
def redirect_node(node):
    node = get_node(node)
    return flask.redirect(flask.url_for(
        '.node', node=node['name'], environment=node['chef_environment']))


@ui.route('/<environment>/nodes/<string:node>')
def node(node):
    return flask.render_template('node.html', node=get_node(node, [
        'run_list',
        'role',
        'recipes',
        'packages'
    ]))


# Packages

@ui.route('/packages/<string:type>/<string:name>')
@ui.route('/<environment>/packages/<string:type>/<string:name>')
def package(type, name):
    nodes = flask.current_app.chef.partial_search('node', {
        'chef_environment': flask.g.chef_environment,
        'packages_' + type: name
    }, {
        'package_version': ['packages', type, name, 'version']
    })
    return flask.render_template(
        'package.html', package_type=type, package_name=name, nodes=nodes)
